import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.List;
import java.awt.geom.*;
import java.util.ArrayList;

/**
 * A standard Object in the game that is pertinent to the game world. This excludes buttons, cursors, etc.
 * 
 * @author Younghwan Choi
 * @version boo!
 */
public class DestructibleObjects  extends IngameSprite
{
    /**
     * stores the value if the unit stepped on a spike
     */
    public boolean spiked;
    protected int battledur;
    /**
     * Maximum hit points of the object.
     */
    protected int maxhitpoints;
    /**
     * The team that this object belongs to. 6 is neutral. (up to 5 teams, with 3,4,5 used only in campaigns, if it is implemented);
     */
    public int side;
    /**
     * The armor type of the unit. Determines the damage calculation.
     * Armor type: Light-1, Medium-2, Heavy-3, Structure-4
     */
    protected int armortype;
    /**
     * the quantity of armor. Reduces the initial damage (before the armor type factoring) by the amount.
     */
    protected int armorquantity;
    /**
     * The range. Used in the attack parameter.
     */
    protected int range;
    /**
     * the actual hit points of the object.
     */
    public double hitpoints;
    /**
     * The weapon type of the unit. Determines damage calculation.
     * Weapon type: AP-1, AT-2, Non-Artillery Explosive-3, Flame-4, Chemical-5, Tank Destroyer-6, Normal-7, Artillery-8;
     */
    protected int weapontype;
    /**
     * The damage amount before reduction and factoring.
     */
    protected int damage;
    /**
     * The value which is used to determine reload time after the object attacks.
     */
    protected int delay;
    /**
     * The integer value that holds the time remaining before the object can attack again.
     */
    protected int reload = 0;
    /**
     * Checks if this unit is a flyer.
     */
    protected boolean flyer;
    /**
     * Checks if the unit can hit ground units.
     */
    protected boolean canhitground;
    /**
     * Checks if the unit can hit air units.
     */
    protected boolean canhitair;
    /**
     * The speed of the unit.
     */
    protected int speed;
    /**
     * The array list of buffs. Buffs provide different bonuses and penalties for the object in question.
     */
    protected ArrayList<Buff> buffs = new ArrayList<Buff>();
    /**
     * Stores if the object have the health meter. (And progress bar for some units/structures)
     */
    protected boolean hasmeter;
    /**
     * Stores the value should the unit have the health meter.
     */
    protected boolean meteron;
    /**
     * Selected? Enough said.
     */
    protected boolean selected;
    /**
     * Percentage, found by hitpoints/maxhitpoints
     */
    protected double hitpercent;
    /**
     * The health meter.
     */
    protected HitPointCheck hpc;
    /**
     * ID of action.
     */
    protected int unitaction;
    /**
     * The player object that owns the object.
     */
    protected Player owner;
    /**
     * The frame (or the image) ID.
     */
    protected int frame;
    /**
     * The int value, after it reaches the framechangerate (static) value, then frame changes.
     */
    protected int framechange;
    /**
     * Checks if the image must be flipped
     */
    protected boolean mirror;
    /**
     * Cost of the object
     */
    protected int cost;
    /**
     * Vertical Translation
     */
    protected int verttrans;
    /**
     * Alignment (for GreenfootImages)
     */
    public boolean changealignment;
    /**
     * stores the boolean value: returns true if the AI (if it is owned by an AI player) has added the object into account.
     */
    public boolean AIcount;
    /**
     * The target that the object has. If the target is not null, it will attack.
     */
    protected DestructibleObjects target;
    /**
     * If the owner of the object is an AI player, holds the AI.
     */
    public AI ai;
    /**
     * Identification Number of the unit
     */
    public int id;
    /**
     * Applies to certain units only.
     */
    public int dir;
    /**
     * For arrays to store. This is used to see if only one type of unit is selected.
     */
    public boolean arraycount;
        /**
     * Stores the previous command when the unit attacks
     */
    protected int prevact;
    /**
     * Row position. Used by AI.
     */
    private int rowpos;
    public DestructibleObjects(boolean dummy, int ix, int iy)
    {
        if(dummy)
        {
            maxhitpoints = 50;
            hitpoints = Greenfoot.getRandomNumber(50)+1;
            side = 2;
            armortype = 3;
            armorquantity = 2;
            weapontype = 3;
            damage = 5;
            delay = 7;
            canhitground = true;
            canhitair = false;
            flyer = false;
            
            speed = 0;
            range = 50;
            setImage("Dummy.png");
            mapX = ix;
            mapY = iy;
        }
    }
    public DestructibleObjects(int iniside)
    {
        side = iniside;
    }
    public DestructibleObjects()
    {
    }
    public void act() 
    {
        // Add your action code here.
        createHealthBar();
        perish();
    }    
    /**
     * sets Side. Unused.
     */
    public void setSide(int ini)
    {
        side = ini;
    }
    /**
     * The collision box
     */
    public Rectangle2D.Double collisionBox()
    {
        return new Rectangle2D.Double(realX() - getImage().getWidth()/2, realY() - getImage().getHeight()/2, getImage().getWidth(), getImage().getHeight());
    }
    /**
     * The Attack Parameter: it is an abstract rectangle that determines if the unit should attack. Some units have special attack parameters and overrides it.
     */
    public Rectangle2D.Double attackParameter()
    {
        Rectangle2D.Double a = new Rectangle2D.Double(0,0,0,0);
        if(side == 1)
        {
            if(flyer)
            {
                a = new Rectangle2D.Double(getX(), getY() + height/2, range, 60+height);
                if(this instanceof NapalmBomber)
                {
                    a = new Rectangle2D.Double(getX(), getY() + height/2, range, 30+height);
                }
            }
            else
            {
                a = new Rectangle2D.Double(getX(), getY() - 30, range, 60);
            }
        }
        else
        {
            if(flyer)
            {
                a = new Rectangle2D.Double(getX()-range, getY() - 30 + height, range, 60);
            }
            else
            {
                a = new Rectangle2D.Double(getX()-range, getY() - 30, range, 60);
            }
        }
        return a;
    }
    /**
     * Attack Parameter Flank: A backward version of the attack parameter. Flanking (by sliding units vertically to face the enemy or attacking backwards) is rewarding.
     */
    public Rectangle2D.Double attackParameterFlank()
    {
        Rectangle2D.Double a = new Rectangle2D.Double(0,0,0,0);
        if(side == 2)
        {
            if(flyer)
            {
                a = new Rectangle2D.Double(getX(), getY() + height/2, range, 60+height);
                if(this instanceof NapalmBomber)
                {
                    a = new Rectangle2D.Double(getX(), getY() + height/2, range, 30+height);
                }
            }
            else
            {
                a = new Rectangle2D.Double(getX(), getY() - 30, range, 60);
            }
        }
        else
        {
            if(flyer)
            {
                a = new Rectangle2D.Double(getX()-range, getY() - 30 + height, range, 60);
            }
            else
            {
                a = new Rectangle2D.Double(getX()-range, getY() - 30, range, 60);
            }
        }
        return a;
    }
    /**
     * Removes object if the unit has 0 or less hit points.
     */
    public void perish()
    {
        if(battledur > 0)
        {
            battledur --;
        }
        if(hitpoints <= 0)
        {
            if(this instanceof Units)
            {
                Units a = (Units) this;
                //Units that are some distance from the screen will not be heard.
                if(!a.mechanical && a.getX() > - 200 && a.getX() < 1000)
                {
                    switch(Greenfoot.getRandomNumber(7))
                    {
                        case 0: Greenfoot.playSound("DeathFix1.wav"); break;
                        case 1: Greenfoot.playSound("Death2.wav"); break;
                        case 2: Greenfoot.playSound("DeathFix3.wav"); break;
                        case 3: Greenfoot.playSound("DeathFix4.wav"); break;
                        case 4: Greenfoot.playSound("DeathFix5.wav"); break;
                        case 5: Greenfoot.playSound("DeathFix6.wav"); break;
                        case 6: Greenfoot.playSound("DeathFix7.wav"); break;
                    }
                }
                else if(a.getX() > - 200 && a.getX() < 1000)
                {
                    switch(Greenfoot.getRandomNumber(3))
                    {
                        case 0: Greenfoot.playSound("VehicleExplosion.wav"); break;
                        case 1: Greenfoot.playSound("VehicleExplosionFix2.wav"); break;
                        case 2: Greenfoot.playSound("VehicleExplosionFix3.wav"); break;
                    }
                }
                if(ai != null)
                {
                    ai.units[id] --;
                }
            }
            if(this instanceof Structures)
            {
                //Strutures that are way away from the screen will not be heard.
                if(ai != null)
                {
                    ai.Structures[id] --;
                }
            }
            getWorld().removeObject(this);
        }
    }
    
    /**
     * Creates the health bar.
     */
    public void createHealthBar()
    {
        if(hpc == null && meteron)
        {
                hpc = new HitPointCheck(this);
                getWorld().addObject(hpc, getX(), getY());
                hasmeter = true;
        }
        try
        {
        if(hpc.getWorld() == null && meteron)
        {
            getWorld().addObject(hpc, getX(), getY());
            hasmeter = true;
        }
    }
    catch(NullPointerException e)
    {
    }
            hitpercent = (hitpoints / ((double)maxhitpoints));
            if(Greenfoot.mousePressed(null))
            {
                MouseInfo ms = Greenfoot.getMouseInfo();
                if(ms.getButton() == 3)
                {
                    meteron = false;
                }
            }            
    }
    /**
     * Deselects the object, if it is selected.
     */
    public void deselect()
    {
        MouseInfo ms = Greenfoot.getMouseInfo();
        if(Greenfoot.mousePressed(null))
        {
            if(ms.getButton() == 3)
            {
                selected = false;
            }
        }
    }
    /**
     * returns the player object that owns the object.
     */
    public Player getOwner()
    {
        return owner;
    }
    /**
     * Take damage: Factors are listed, and uses the armor and weapon types to determine damage. If the damage is less than or equal to the armor, it does half of a point
     * of damage. Also, if an unit is attacked, the unit will try to move forwards to attack back. (This makes it even more beneficial to attack from the back)
     */
    public void takedamage(int damage, int weapontype)
    {
        if(getAction() != 1 && getAction() != 3 && getAction() != 2 && getTrans() <= 0 && !(this instanceof Structures) && !(this instanceof WorkerSphere) && !(this instanceof WorkerBaby) && !(this instanceof Harvester))
        {
            setBehavior(1);
        }
        if(battledur <= 1)
        {
            battledur += damage * 30;
        }
        if(genmaches.bd == null || genmaches.bd.getWorld() == null)
        {
            genmaches.bd = new BattleDummy(realX(), realY());
            getWorld().addObject(genmaches.bd, getX(), getY());
        }
        if(armortype == 1)
        {
            if(armorquantity >= damage)
            {
                hitpoints -= .5;
            }
            else
            {
            switch(weapontype)
            {
                case 1: hitpoints -= damage * 1.5; break;
                case 2: hitpoints -= damage * .25; break;
                case 3: hitpoints -= damage * 1.0; break;
                case 4: hitpoints -= damage * 1.5; break;
                case 5: hitpoints -= damage * 1.5; break;
                case 6: hitpoints -= damage * .15; break;
                case 7: hitpoints -= damage * 1.0; break;
                case 8: hitpoints -= damage * .75; break;
            }
        }
        }
        if(armortype == 2)
        {
            if(armorquantity >= damage)
            {
                hitpoints -= .5;
            }
            else
            {
            switch(weapontype)
            {
                case 1: hitpoints -= damage * .4; break;
                case 2: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 3: hitpoints -= (damage - armorquantity) * 1.5; break;
                case 4: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 5: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 6: hitpoints -= (damage - armorquantity) * .75; break;
                case 7: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 8: hitpoints -= (damage - armorquantity) * 1.0; break;
            }
        }
        }
        if(armortype == 3)
        {
            if(armorquantity >= damage)
            {
                hitpoints -= .5;
            }
            else
            {
            switch(weapontype)
            {
                case 1: hitpoints -= (damage - armorquantity) * .15; break;
                case 2: hitpoints -= (damage - armorquantity) * 2; break;
                case 3: hitpoints -= (damage - armorquantity) * .75; break;
                case 4: hitpoints -= (damage - armorquantity) * .5; break;
                case 5: hitpoints -= (damage - armorquantity) * .5; break;
                case 6: hitpoints -= (damage - armorquantity) * 3; break;
                case 7: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 8: hitpoints -= (damage - armorquantity) * 1.; break;
            }
            }
        }
        if(armortype == 4)
        {
            if(armorquantity >= damage)
            {
                hitpoints -= .5;
            }
            else
            {
            switch(weapontype)
            {
                case 1: hitpoints -= (damage - armorquantity) * .2; break;
                case 2: hitpoints -= (damage - armorquantity) * .5; break;
                case 3: hitpoints -= (damage - armorquantity) * 1.0; break;
                case 4: hitpoints -= (damage - armorquantity) * 1.; break;
                case 5: hitpoints -= (damage - armorquantity) * .1; break;
                case 6: hitpoints -= (damage - armorquantity) * .25; break;
                case 7: hitpoints -= (damage - armorquantity) * .5; break;
                case 8: hitpoints -= (damage - armorquantity) * 2.; break;
            }
        }
        }
    }
    /**
     * Sets behavior.
     */
    public void setBehavior(int u)
    {
        unitaction = u;
    }
    /**
     * gets unit action ID.
     */
    public int getAction()
    {
        return unitaction;
    }
    /**
     * gets the team this object belongs to.
     */
    public int getSide()
    {
        return side;
    }
    /**
     * gets speed.
     */
    public int getSpeed()
    {
        if(spiked && speed - 1 > 0)
        {
            return speed -1;
        }
        else
        {
            return speed;
        }
    }
    /**
     * Sets the real location (or map location)
     */
    public void setRealLocation(int x, int y)
    {
        mapX = x;
        mapY = y;
    }
    /**
     * selects unit. creates meter.
     */
    public void select()
    {
        if(owner == genmaches.thePlayer)
        {
        selected = true;
        }
        meteron = true;
    }
    /**
     * returns the boolean value whether or not the unit is selected.
     */
    public boolean selected()
    {
        return selected;
    }
    /**
     * Sets the amount of vertical translation.
     */
    public void setVerticalTranslation(int i)
    {
        verttrans = i;
    }
    /**
     * gets the vertical translation needed.
     */
    public int getTrans()
    {
        return verttrans;
    }
    /**
     * Adds a buff to the unit.
     */
    public void addBuff(Buff a)
    {
        buffs.add(a);
    }
    /**
     * returns the cooldown of the weapon. used for buff purposes.
     */
    public int getDelay()
    {
        return delay;
    }
    public DestructibleObjects getTarget()
    {
        return target;
    }
    public void setPBehavior(int ini)
    {
        prevact = ini;
    }
    /**
     * Checks for certain upgrades (used mostly by EBC)
     */
    public void checkUpgrades()
    {
        if(owner.Upgrades[3] > 0)
        {
            if(this instanceof PistoleerBaby)
            {
                if(maxhitpoints != 36)
                {
                    hitpoints += 6;
                    maxhitpoints = 36;
                    hitpoints += (int)(6*(hitpoints/(double)maxhitpoints));
                    delay = 16;
                }
            }
            if(this instanceof GrenadierBaby)
            {
                if(maxhitpoints != 54)
                {
                    hitpoints += 9;
                    maxhitpoints = 54;
                    delay = 45;
                    hitpoints += (int)(9*(hitpoints/(double)maxhitpoints));
                }
            }
        }
        if(owner.Upgrades[4] > 0)
        {
            if(this instanceof PistoleerBaby)
            {
                if(delay != 14)
                {
                    delay = 14;
                }
            }
            if(this instanceof GrenadierBaby)
            {
                if(damage != 9)
                {
                    damage = 12;
                }
            }
        }
        if(owner.Upgrades[5] > 0)
        {
            if(this instanceof SuicideBaby)
            {
                if(damage != 50)
                {
                    damage = 50;
                }
            }
        }
        if(owner.Upgrades[6] > 0)
        {
            if(this instanceof TroopCarrier)
            {
                if(delay != 8)
                {
                    delay = 8;
                }
            }
        }
        if(owner.Upgrades[7] > 0)
        {
            if(this instanceof Artillery)
            {
                if(damage != 40)
                {
                    damage = 40;
                }
            }
        }
    }
    /**
     * Returns whether or not the unit is in battle. (Or under attack in the case of structures)
     */
    public boolean underAttack()
    {
        return battledur > 0;
    }
    /**
     * Returns the array list of buff in the object.
     */
    public ArrayList<Buff> getBuffs()
    {
        return buffs;
    }
    public double getHitpoints()
    {
        return hitpoints;
    }
    public int getMaxHitpoints()
    {
        return maxhitpoints;
    }
    public void heal(double am)
    {
        hitpoints += am;
        if(hitpoints > maxhitpoints)
        {
            hitpoints = maxhitpoints;
        }
    }
        public int getRow()
    {
        if(getAction() == 4 && getTrans() > 0)
        {
            return rowpos - 1;
        }
        else if(getAction() == 5 && getTrans() > 0)
        {
            return rowpos + 1;
        }
        else
        {
            return rowpos;
        }
    }
    /**
     * Uses the real Y value to get row.
     */
    public void getRowPosition()
    {
        if(realY() < 0)
        {
            rowpos = 12;
            if(getTrans() <= 0 && this instanceof Units)
            {
                setVerticalTranslation(getImage().getHeight());
                setBehavior(5);
            }
        }
        else if(realY() < 40)
        {
            rowpos = 0;
        }
        else if(realY() >= 40 && realY() < 80)
        {
            rowpos = 1;
        }
        else if(realY() >= 80 && realY() < 120)
        {
            rowpos = 2;
        }
        else if(realY() >= 120 && realY() < 160)
        {
            rowpos = 3;
        }
        else if(realY() >= 160 && realY() < 200)
        {
            rowpos = 4;
        }
        else if(realY() >= 200 && realY() < 240)
        {
            rowpos = 5;
        }
        else if(realY() >= 240 && realY() < 280)
        {
            rowpos = 6;
        }
        else if(realY() >= 280 && realY() < 320)
        {
            rowpos = 7;
        }
        else if(realY() >= 320 && realY() < 360)
        {
            rowpos = 8;
        }
        else if(realY() >= 360 && realY() < 400)
        {
            rowpos = 9;
        }
        else
        {
            rowpos = 10;
        }
    }
}
